"""
General HTTP Client with Pydantic support
"""
import logging
import typing as t

import requests
from pydantic import BaseModel

from taktile_client.config import version

logger = logging.getLogger(__name__)

API_KEY_HEADER = "X-Api-Key"


class API:
    """API.
    Basic HTTP client with support for model parsing
    """

    def __init__(
        self,
        api_base: str,
        api_key: t.Optional[str] = None,
        headers: t.Optional[t.Dict[str, str]] = None,
    ):
        self._base = api_base
        self.set_headers(headers)
        self.set_api_key(api_key)

    def set_headers(self, headers: t.Optional[t.Dict[str, str]]) -> None:
        """set_headers.
        Sets headers. If specified it will override the base headers, if not
        it will use the base headers.

        Parameters
        ----------
        headers: t.Optional[t.Dict[str, str]]
            headers to merge
        """
        base_headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
            "User-Agent": f"taktile-client/{version}",
        }
        if headers:
            self._headers = {**base_headers, **headers}
        else:
            self._headers = base_headers

    def set_api_key(self, api_key: t.Optional[str]) -> None:
        """set_api_key.
        Sets api key header. If specified will use the specified one.  If not
        specified, will use one that is specified in the environment.
        Parameters
        ----------
        api_key : t.Optional[str]
            api_key to be set
        """
        if api_key:
            self._headers[API_KEY_HEADER] = api_key
        else:
            self._headers.pop(API_KEY_HEADER, None)

    def call(
        self,
        *,
        verb: str,
        path: str,
        model: t.Optional[t.Type[BaseModel]] = None,
        raw: bool = False,
        timeout: t.Optional[float] = None,
        **kwargs: t.Any,
    ) -> t.Any:
        """
        Call an endpoint

        Parameters
        ----------
        verb : str
            verb to use
        path : str
            path to call
        model : t.Optional[t.Type[BaseModel]]
            model to parse the response into
        raw : bool
            if true, return the unparsed response
        kwargs :
            kwargs passed on to requests.<verb>()
        """
        path = requests.compat.urljoin(self._base, path)
        logger.debug("%s %s: %s", verb, path, str(kwargs))
        response = getattr(requests, verb)(
            path,
            timeout=timeout,
            **kwargs,
            headers=self._headers,
        )

        if raw:
            return response

        if not response.ok:
            logger.error(
                "%s %s: %s - %s - %s",
                verb,
                path,
                str(kwargs),
                response.status_code,
                response.content,
            )
            response.raise_for_status()

        logger.debug(
            "%s %s: %s - %s - %s",
            verb,
            path,
            str(kwargs),
            response.status_code,
            response.content,
        )

        json = response.json()

        if model is None:
            return json

        if isinstance(json, list):
            return [model(**x) for x in json]

        return model(**json)



"""
API object to talk to a model
"""
import typing as t

from taktile_types.schemas.deployment.endpoint import InfoEndpointResponseModel

from taktile_client.http_client import API


class Model(API):
    """Model api object"""

    def __init__(
        self,
        url: str,
        api_key: t.Optional[str],
        headers: t.Optional[t.Dict[str, str]] = None,
    ):
        super().__init__(api_base=url, api_key=api_key, headers=headers)

    def get_info(self) -> InfoEndpointResponseModel:
        """Retrieve model info

        Returns
        -------
        InfoEndpointResponseModel
            info endpoint response model
        """
        return t.cast(
            InfoEndpointResponseModel,
            self.call(
                verb="get", path="/info", model=InfoEndpointResponseModel
            ),
        )



"""
Exceptions used in Taktile Client
"""


class TaktileClientException(Exception):
    """A generic taktile client exception"""


class HTTPException(TaktileClientException):
    """A Exception that is the result from an HTTP Exception"""

    def __init__(self, status_code: int, detail: str) -> None:
        super().__init__()
        self._status_code = status_code
        self._detail = detail

    def __repr__(self) -> str:
        class_name = self.__class__.__name__
        return (
            f"{class_name}"
            f"(status_code={self._status_code!r}, detail={self._detail!r})"
        )


class APIClientExceptionRetryable(HTTPException):
    """A retryable client api exception"""


class APIClientException(HTTPException):
    """A generic client api exception"""


class SerializationError(TaktileClientException):
    """A serialization exception"""


class IncompatibleVersionError(TaktileClientException):
    """Client server version is not compatible"""



"""
Numpy rest serialization
"""
import json
import typing as t

import numpy as np
import numpy.typing as npt

from .utils import JSONStructure


def numpy_deserialize(
    *, value: JSONStructure, sample: npt.NDArray[t.Any]
) -> npt.NDArray[t.Any]:
    """
    Deserialize a numpy array

    Parameters
    ----------
    value : JSONStructure
        value to deserialize
    sample : npt.NDArray[t.Any]
        sample numpy array

    Returns
    -------
    npt.NDArray[t.Any]

    """
    del sample
    return np.array(value, dtype=object)


def numpy_serialize(*, value: npt.NDArray[t.Any]) -> str:
    """
    Serialize a numpy array

    Parameters
    ----------
    value : npt.NDArray[t.Any]
        value to serialize

    Returns
    -------
    str
        serialized numpy array
    """
    return json.dumps(value.tolist())


def numpy_to_example(*, value: npt.NDArray[t.Any]) -> str:
    """
    Produce an example for a numpy array

    Parameters
    ----------
    value : npt.NDArray[t.Any]
        numpy array to produce an example from

    Returns
    -------
    str
        serialized example
    """
    return json.dumps(value[0:1].tolist())



"""
API client to talk to deployment-api
"""
import typing as t
import urllib.parse

from taktile_types.schemas.deployment.model import Model, ModelDeployment

from taktile_client.config import settings
from taktile_client.exceptions import HTTPException, TaktileClientException
from taktile_client.http_client import API


class DeploymentApiClient(API):
    """deployment-api client object"""

    def __init__(
        self, api_key: t.Optional[str], headers: t.Optional[t.Dict[str, str]]
    ):
        super().__init__(settings.DEPLOYMENT_API_URL, api_key, headers)

    def get_models(self) -> t.List[Model]:
        """Get all models for a user

        Returns
        -------
        t.List[Model]
            List of models
        """
        return t.cast(
            t.List[Model],
            self.call(verb="get", path="/api/v1/models", model=Model),
        )

    def get_model_deployment_by_repo_full_name_and_branch(
        self, principal: str, repository: str, branch: str
    ) -> ModelDeployment:
        """Get a specific model deployment

        Parameters
        ----------
        principal : str
            principal to look for
        repository : str
            repository to look for
        branch : str
            branch to look for. Can be either "main" or "refs/heads/main"

        Raises
        ------
        TaktileClientException
            raised if model deployment is not found

        Returns
        -------
        ModelDeployment
            the deployment requested
        """
        try:
            params = {
                "principal": principal,
                "repository": repository,
                "branch": branch,
            }
            url = (
                f"/api/v1/models/deployment/?{urllib.parse.urlencode(params)}"
            )
            model_deployment = t.cast(
                ModelDeployment,
                self.call(verb="get", path=url, model=ModelDeployment),
            )
            return model_deployment
        except HTTPException as err:
            raise TaktileClientException(
                f"Deployment {principal} - {repository} - {branch} not found"
            ) from err

    def get_host(self, principal: str, repository: str, branch: str) -> str:
        """Get the host of a deployment

        Parameters
        ----------
        principal : str
            principal to look for
        repository : str
            repository to look for
        branch : str
            branch to look for. Can be either "main" or "refs/heads/main"

        Raises
        ------
        TaktileClientException
            raised if model deployment is not found

        Returns
        -------
        str
            the host of the deployment
        """
        deployment = self.get_model_deployment_by_repo_full_name_and_branch(
            principal, repository, branch
        )

        if deployment.public_docs_url is None:
            raise TaktileClientException(
                f"Deployment {principal} - {repository} - {branch} "
                "is not ready to be called"
            )

        return deployment.public_docs_url



"""
Series rest serialization
"""
import typing as t

import pandas as pd  # type: ignore

from .utils import JSONStructure


def series_deserialize(
    *, value: JSONStructure, sample: pd.Series
) -> pd.Series:
    """
    Deserialize a pandas Series

    Parameters
    ----------
    value : JSONStructure
        value to deserialize
    sample : pd.Series
        sample series

    Returns
    -------
    pd.Series
        deserialized dataframe
    """
    del sample
    return pd.Series(value)


def series_serialize(*, value: pd.Series) -> str:
    """
    Serialize pandas series

    Parameters
    ----------
    value : pd.Series
        value to serialize

    Returns
    -------
    str
        serialized series
    """
    return t.cast("str", value.to_json(orient="records", date_format="iso"))


def series_to_example(*, value: pd.Series) -> str:
    """
    Produce an example for a pandas series

    Parameters
    ----------
    value : pd.Series
        series to produce an example for

    Returns
    -------
    str
        json serialized example
    """
    return t.cast(
        "str", value.iloc[[0]].to_json(orient="records", date_format="iso")
    )



"""
Client for talking to arrow endpoints
"""
import logging
import typing as t

# pylint: disable-next=no-name-in-module
from pyarrow._flight import (  # type: ignore
    ClientAuthReader,
    ClientAuthSender,
    FlightServerError,
)
from pyarrow.flight import (  # type: ignore
    ClientAuthHandler,
    FlightClient,
    Ticket,
)
from taktile_types.schemas.deployment.endpoint import (
    InfoEndpointResponseModelFlight,
)

from taktile_client.arrow.serialize import deserialize
from taktile_client.arrow.utils import load_certs
from taktile_client.client import Client
from taktile_client.config import version as cversion
from taktile_client.version import assert_supports_version

from ..config import settings
from .endpoint import ArrowEndpoints

logger = logging.getLogger(__name__)


class ApiKeyClientAuthHandler(ClientAuthHandler):  # type: ignore
    """Auth Handler"""

    def __init__(self, api_key: str):
        super().__init__()
        self._api_key = api_key

    def authenticate(
        self, outgoing: ClientAuthSender, incoming: ClientAuthReader
    ) -> None:
        """authenticate"""
        outgoing.write(self._api_key)
        self._api_key = incoming.read()

    def get_token(self) -> str:
        """get_token"""
        return self._api_key


class ArrowClient(Client[ArrowEndpoints]):
    """
    Arrow client for talking to models deployed with Taktile.
    """

    def _url_from_host(self, host: str) -> str:
        if settings.DEBUG:
            return f"grpc://{host}:5005"
        return f"grpc+tls://{host}:5005"

    def _initalize_endpoints(
        self, url: str, api_key: t.Optional[str]
    ) -> ArrowEndpoints:

        certs = load_certs()
        logger.debug("Performing authentication request against %s", url)
        client = FlightClient(location=url, tls_root_certs=certs)
        if api_key:
            client.authenticate(ApiKeyClientAuthHandler(api_key=api_key))
        actions = client.list_actions()
        self.info = _get_info(client)
        if self.info:
            assert_supports_version(
                client_version=cversion,
                server_version=self.info.taktile_cli,
            )
        return ArrowEndpoints(client=client, actions=actions)


def _get_info(
    client: FlightClient,
) -> t.Union[InfoEndpointResponseModelFlight, None]:
    try:
        return InfoEndpointResponseModelFlight(
            **deserialize(client.do_get(Ticket("__info")).read_all())
        )
    except FlightServerError as err:
        logger.warning("Failed to get info %s", repr(err))
        return None



"""
Endpoint implementation for the rest client.
"""
import json
import typing as t

from taktile_types.enums.endpoint import ArrowFormatKinds
from taktile_types.schemas.deployment.endpoint import EndpointInfoSchema

from taktile_client.config import arrow_available
from taktile_client.http_client import API
from taktile_client.rest.serialize import deserialize, serialize
from taktile_client.rest.utils import is_serializable, safe_post

if arrow_available:
    import numpy as np
    import pandas as pd  # type: ignore


class ClientRestEndpoint:
    """
    The actual endpoint object.

    This is a Rest endpoint, exposing function calling through `__call__` and
    explainer endpoints through `explain`. This is the object that is
    `client.endpoints.repayment`.

    Parameters
    ----------
    api : API
        the api to use for making the calls
    info : EndpointInfoSchema
        the endpoint info of this endpoint
    """

    def __init__(self, *, api: API, info: EndpointInfoSchema):
        self._api = api
        self._info = info
        self.__name__ = info.name
        self._y: t.Any = None

    def __call__(
        self, payload: t.Any, *, retries: int = 3, timeout: float = 10.0
    ) -> t.Any:
        """
        Call the endpoints

        Parameters
        ----------
        payload : t.Any
            the payload
        retries : int
            how often to retry on recoverable errors
        timeout : float
            timeout of call

        Returns
        -------
        t.Any
            The response of the endpoint
        """
        serializable = is_serializable(payload)
        if serializable:
            data = serialize(payload)
        else:
            data = json.dumps(payload)
        result = safe_post(
            api=self._api,
            url=self._info.path,
            retries=retries,
            timeout=timeout,
            data=data,
        )
        if serializable:
            if self._info.response_kind == ArrowFormatKinds.DATAFRAME:
                return deserialize(pd.DataFrame(), value=result)
            if self._info.response_kind == ArrowFormatKinds.SERIES:
                return deserialize(pd.Series(), value=result)
            if self._info.response_kind == ArrowFormatKinds.ARRAY:
                return deserialize(np.array([]), value=result)
            return result
        return result

    def explain(
        self, payload: t.Any, *, retries: int = 3, timeout: float = 10.0
    ) -> t.Any:
        """
        Call the endpoint's explainer endpoint.

        Parameters
        ----------
        payload : t.Any
            the payload
        retries : int
            how often to retry on recoverable errors
        timeout : float
            timeout of call

        Returns
        -------
        t.Any
            The response of the explainer
        """
        if not self._info.explain_path:
            raise ValueError(
                "The endpoint {self._info.name} is of type {self._info.kind}"
                " and does not have an explain endpoint"
            )
        if is_serializable(payload):
            data = serialize(payload)
        else:
            data = json.dumps(payload)
        return safe_post(
            api=self._api,
            url=self._info.explain_path,
            retries=retries,
            timeout=timeout,
            data=data,
        )

    def X(self) -> t.Any:  # pylint: disable=invalid-name
        """Retrieve sample data

        Returns
        -------
        t.Any
            sample data to call the endpoint
        """
        return self._info.input_example

    def y(self) -> t.Any:  # pylint: disable=invalid-name
        """Retrieve sample response

        Returns
        -------
        t.Any
            sample response from the endpoint
        """
        # Use functools.cache when upgrading to Python 3.8
        if not self._y:
            self._y = safe_post(
                api=self._api,
                url=self._info.path,
                json=self._info.input_example,
            )
        return self._y


class RestEndpoints:  # pylint: disable=too-few-public-methods
    """
    A collection object for Rest endpoints.

    This is the `endpoints` object on Taktile Rest Clients.

    Parameters
    ----------
    api : API
        the api to use for calls
    infos : t.List[EndpointInfoSchema]
        a list of endpoints to instantiate
    """

    def __init__(self, *, api: API, infos: t.List[EndpointInfoSchema]):

        for info in infos:
            setattr(self, info.name, ClientRestEndpoint(api=api, info=info))



"""
Dataframe rest serialization
"""
import typing as t

import pandas as pd  # type: ignore

from .utils import JSONStructure


def dataframe_deserialize(
    *, value: JSONStructure, sample: pd.DataFrame
) -> pd.DataFrame:
    """
    Deserialize a dataframe

    Parameters
    ----------
    value : JSONStructure
        value to deserialize
    sample : pd.DataFrame
        sample dataframe

    Returns
    -------
    pd.DataFrame
        deserialized dataframe
    """
    del sample
    return pd.DataFrame(value)


def dataframe_serialize(*, value: pd.DataFrame) -> str:
    """
    Serialize a dataframe

    Parameters
    ----------
    value : pd.DataFrame
        value to serialize

    Returns
    -------
    str
        serialized dataframe
    """
    return t.cast("str", value.to_json(orient="records", date_format="iso"))


def dataframe_to_example(*, value: pd.DataFrame) -> str:
    """
    Produce an example for a dataframe

    Parameters
    ----------
    value : pd.DataFrame
        dataframe to produce an example from

    Returns
    -------
    str
        json serialized example
    """
    return t.cast(
        "str", value.iloc[[0]].to_json(orient="records", date_format="iso")
    )



"""
Utilities for semantic versioning.
"""

from packaging import version

from taktile_client.config import settings
from taktile_client.exceptions import IncompatibleVersionError


def assert_supports_version(
    *, client_version: str, server_version: str
) -> None:
    """assert_supports_version checks if two version are compatible. Raises if not

    Parameters
    ----------
    client_version : str
        client version
    server_version : str
        server version

    Raises
    ------
    IncompatibleVersionError
        If the two versions are not compatible. Error includes explanation
    """
    if settings.DEBUG:
        return

    cversion = version.Version(client_version)
    sversion = version.Version(server_version)

    supported_major_versions = {cversion.major, cversion.major - 1}

    if sversion.major not in supported_major_versions:
        raise IncompatibleVersionError(
            f"Client version {cversion} does not "
            f"support server version {sversion}. Please update."
        )

    if sversion.major == 0 and sversion.minor != 8:
        raise IncompatibleVersionError(
            f"Client version {cversion} does not "
            f"support server version {sversion}. Please update."
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['taktile_client',
 'taktile_client.arrow',
 'taktile_client.arrow.serialize',
 'taktile_client.rest',
 'taktile_client.rest.serialize']

package_data = \
{'': ['*']}

install_requires = \
['packaging>=23.0,<24.0',
 'pydantic>=1.8.2,<2.0.0',
 'requests>=2.26.0,<3.0.0',
 'setuptools>=0.0.0',
 'taktile-types>=0.17.0,<1.0.0',
 'tenacity>=8.0.1,<9.0.0']

extras_require = \
{'arrow': ['numpy>=1.23,<2.0',
           'pandas>=1.4,<2.0',
           'pyarrow>=8,<=11',
           'certifi==2021.10.8']}

setup_kwargs = {
    'name': 'taktile-client',
    'version': '1.0.7',
    'description': 'A lightweight client to call models deployed on Taktile.',
    'long_description': '# Taktile Client\n\n[![pypi status](https://img.shields.io/pypi/v/taktile-client.svg)](https://pypi.python.org/pypi/taktile-client)\n[![License](https://img.shields.io/badge/License-Apache%202.0-blue.svg)](https://www.apache.org/licenses/LICENSE-2.0)\n\nTaktile enables data science teams to industrialize, scale, and maintain machine learning models. Our ML development platform makes it easy to create your own end-to-end ML applications:\n\n- Turn models into auto-scaling APIs in a few lines of code\n- Easily add model tests\n- Create and share model explanations through the Taktile UI\n\n`taktile-client` is a stand-alone python client which can be used to make requests to Taktile deployments via REST or [Arrow Flight](https://arrow.apache.org/docs/format/Flight.html). If you require the full Taktile dev tooling, consider installing [taktile-cli](https://pypi.org/project/taktile-cli/) instead. Find more information in our [docs](https://docs.taktile.com).\n\nTo install the REST client only, run `pip install taktile-client`. For both REST and Arrow, run `pip install taktile-client\\[arrow]`.\n',
    'author': 'Taktile GmbH',
    'author_email': 'devops@taktile.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""
Utility functions for serialization
"""
import functools
import json
from io import BytesIO
from typing import Any, Callable, Dict, List, Optional, Sequence, Type, Union

import pyarrow  # type: ignore
from pyarrow import json as j


def _serialize_dicts_or_list_of_dicts(
    values: Union[Sequence[Dict[Any, Any]], Dict[Any, Any]]
) -> pyarrow.Table:
    mem_file = BytesIO()
    if isinstance(values, dict):
        values = [values]
    for item in values:
        mem_file.write(json.dumps(item).encode())
    mem_file.seek(0)
    return j.read_json(mem_file)


def _get_final_serializer(
    initial_type: Type[Any], final_type: Type[Any]
) -> Optional[str]:
    if initial_type != final_type:
        if initial_type == list:
            return "NATIVE_list"
        if initial_type == tuple:
            return "FINAL_tuple"
        if initial_type == dict:
            return "FINAL_first"
        raise ValueError()
    return None


def add_deserializing_metadata(
    table: pyarrow.Table,
    *,
    deserializers: Optional[List[str]] = None,
    serializing_cls: Optional[str] = None
) -> pyarrow.Table:
    """add deserializing metadata"""
    metadata: Dict[str, Union[str, bytes]] = {}
    if deserializers:
        metadata["DESERIALIZERS"] = json.dumps(deserializers).encode()
    if serializing_cls:
        metadata["CLS"] = serializing_cls
    return _replace_table_with_metadata(table=table, metadata=metadata)


def _replace_table_with_metadata(
    table: pyarrow.Table, metadata: Dict[Any, Any]
) -> pyarrow.Table:
    final = (
        {**table.schema.metadata, **metadata}
        if table.schema.metadata
        else metadata
    )
    return table.replace_schema_metadata(metadata=final)


def inject_metadata(serializer: Callable[..., Any]) -> Callable[..., Any]:
    """inject metadata into serializer"""

    @functools.wraps(serializer)
    def wraps(cls: Any, *args: Any, **kwargs: Any) -> Any:
        table: pyarrow.Table = serializer(cls, *args, **kwargs)
        table = add_deserializing_metadata(
            table=table, serializing_cls=cls.__name__
        )
        return table

    return wraps



"""
Utilities for the rest client
"""
import logging
import typing as t

import pydantic
import requests
import tenacity

from taktile_client.config import arrow_available
from taktile_client.exceptions import (
    APIClientException,
    APIClientExceptionRetryable,
)
from taktile_client.http_client import API

if arrow_available:
    import numpy as np
    import pandas as pd  # type: ignore

logger = logging.getLogger(__name__)

INTERNAL_SERVER_ERROR = (
    requests.codes.INTERNAL_SERVER_ERROR  # pylint: disable=no-member
)
FORBIDDEN = requests.codes.FORBIDDEN  # pylint: disable=no-member
UNAUTHORIZED = requests.codes.UNAUTHORIZED  # pylint: disable=no-member
UNPROCESSABLE_ENTITY = (
    requests.codes.UNPROCESSABLE_ENTITY  # pylint: disable=no-member
)

DO_NOT_RETRY_STATUS_CODES = [
    INTERNAL_SERVER_ERROR,
    FORBIDDEN,
    UNAUTHORIZED,
    UNPROCESSABLE_ENTITY,
]


def safe_post(
    *,
    api: API,
    url: str,
    json: t.Optional[t.Any] = None,
    data: t.Optional[t.Any] = None,
    retries: int = 3,
    timeout: float = 30.0,
) -> t.Any:
    """HTTP POST request with retrying

    Parameters
    ----------
    api : API
        the underlying api
    url : str
        url to make the request to
    json: t.Any
        data to be sent
    data: t.Any
        serialized data to be sent
    retries : int
        retries
    timeout : float
        timeout
    """

    def my_after(retry_state: t.Any) -> None:
        logger.warning(
            "Timeout while trying to call endpoint on try #%d/%d: %s",
            retry_state.attempt_number,
            retries,
            retry_state.outcome,
        )

    def my_stop(retry_state: t.Any) -> bool:
        if retry_state.attempt_number >= retries:
            logger.error(
                "Giving up trying to call endpoint after %d attempts",
                retry_state.attempt_number,
            )
            return True
        return False

    @tenacity.retry(
        stop=my_stop,
        retry=tenacity.retry_if_exception_type(  # type: ignore
            exception_types=(APIClientExceptionRetryable,)
        ),
        wait=tenacity.wait_random(min=0, max=1),  # type: ignore
        after=my_after,
        reraise=True,
    )
    def wrapped() -> t.Any:
        try:
            response = api.call(
                verb="post",
                path=url,
                raw=True,
                json=json,
                data=data,
                timeout=timeout,
            )
        except (requests.Timeout, requests.ConnectionError) as err:
            raise APIClientExceptionRetryable(
                status_code=-1, detail="A connection error has occured"
            ) from err

        if response.status_code in DO_NOT_RETRY_STATUS_CODES:
            if response.status_code == INTERNAL_SERVER_ERROR:
                raise APIClientException(
                    status_code=response.status_code,
                    detail="An error occurred with your request. "
                    "Check the deployment's logs for more information",
                )
            if response.status_code == UNPROCESSABLE_ENTITY:
                json_response = response.json()
                raise APIClientException(
                    status_code=response.status_code,
                    detail="Unprocessable request body: "
                    f"{json_response['detail']})",
                )

            json_response = response.json()
            raise APIClientException(
                status_code=response.status_code,
                detail=f"Authentication Error: {json_response['detail']}",
            )
        try:
            response.raise_for_status()
        except requests.exceptions.RequestException as err:
            try:
                json_response = response.json()
                raise APIClientExceptionRetryable(
                    status_code=response.status_code,
                    detail=json_response["detail"],
                ) from err
            except (ValueError, KeyError) as err_2:
                # Note that we use ValueError instead of JSONDecodeError above
                # on purpose, because depending on runtime configuration
                # `requests` might choose to use `simplejson` which has a
                # different JSONDecodeError. However both derive from
                # ValueError.
                content = response.text
                raise APIClientExceptionRetryable(
                    status_code=response.status_code,
                    detail=f"An error occurred with your request: {content})",
                ) from err_2
        return response.json()

    return wrapped()


def is_serializable(payload: t.Any) -> bool:
    """Utility Method to check if payload can be serialized

    Parameters
    ----------
    payload : t.Any
        payload Object
    """
    if arrow_available and isinstance(
        payload, (pd.DataFrame, pd.Series, np.ndarray)
    ):
        serializable = True
    elif isinstance(payload, pydantic.BaseModel):
        serializable = True
    else:
        serializable = False
    return serializable



"""
Client for talking to rest endpoints
"""
import typing as t

from taktile_client.client import Client
from taktile_client.config import settings
from taktile_client.config import version as cversion
from taktile_client.http_client import API
from taktile_client.model import Model
from taktile_client.rest.endpoint import RestEndpoints
from taktile_client.version import assert_supports_version


class RestClient(Client[RestEndpoints]):
    """
    Rest client for talking to models deployed with Taktile.
    """

    def _url_from_host(self, host: str) -> str:
        if settings.DEBUG:
            return f"http://{host}"
        return f"https://{host}"

    def _initalize_endpoints(
        self, url: str, api_key: t.Optional[str]
    ) -> RestEndpoints:
        info = Model(url=url, api_key=api_key).get_info()

        api = API(api_base=url, api_key=api_key)
        assert_supports_version(
            client_version=cversion, server_version=info.taktile_cli
        )
        return RestEndpoints(api=api, infos=info.endpoints)
